package  
{
	import flash.geom.Point;
	import net.flashpunk.Entity;
	import net.flashpunk.graphics.Image;
	import net.flashpunk.FP;
	import net.flashpunk.tweens.misc.Alarm;
	
	/**
	 * Base class for planets, asteroids, stars
	 * @author Switchbreak
	 */
	public class CelestialBody extends Entity 
	{
		// Constants
		static private const GRAVITATIONAL_CONSTANT:Number = 75000;
		static private const TRACER_DURATION:Number = 0.5;
		static private const TRACER_RADIUS:uint = 5;
		static private const ZOOM:Number = 0.25;
		
		// Private variables
		public var xVelocity:Number = 0;
		public var yVelocity:Number = 0;
		public var xAcceleration:Number = 0;
		public var yAcceleration:Number = 0;
		public var xPosition:Number = 0;
		public var yPosition:Number = 0;
		public var radius:Number;
		public var moving:Boolean = true;
		public var affectedByPlanets:Boolean = false;
		public var exertsGravity:Boolean = true;
		private var spriteScale:Number;
		
		// Graphics
		private var sprite:Image;
		
		// Objects
		private var leaveTracer:Alarm;
		
		/**
		 * Create and initialize the celestial body
		 *
		 * @param	distance
		 * Starting distance of the body from the center
		 * @param	setRadius
		 * Size of the body
		 * @param	setSprite
		 * Sprite for the planet
		 * @param	setAngle
		 * The angle of rotation to start at
		 */
		public function CelestialBody(distance:Number, setRadius:Number, setSprite:Image, setAngle:Number, setScale:Number = 1)
		{
			// Set object values
			sprite = setSprite;
			radius = setRadius;
			spriteScale = setScale;
			
			if ( distance > 0 )
			{
				CircularVelocity( distance, setAngle );
			}
			
			// Create object
			sprite.scale = GameWorld.zoom;
			sprite.centerOrigin();
			super(xPosition * GameWorld.zoom, yPosition * GameWorld.zoom, sprite);
			
			// Turn on orbital tracing
			//leaveTracer = new Alarm( TRACER_DURATION, LeaveTracer, LOOPING );
			//addTween( leaveTracer, true );
		}
		
		/**
		 * Handle planet movement
		 */
		override public function update():void
		{
			if ( moving )
			{
				// Verlet integration
				var deltaTime:Number = FP.elapsed * GameWorld.timeSpeed;
				var halfDT:Number = deltaTime / 2;
				
				xPosition += xVelocity * deltaTime + xAcceleration * halfDT * deltaTime;
				yPosition += yVelocity * deltaTime + yAcceleration * halfDT * deltaTime;
				
				xVelocity += xAcceleration * halfDT;
				yVelocity += yAcceleration * halfDT;
				
				// Calculate current acceleration due to gravity
				CalculateGravity();
				
				xVelocity += xAcceleration * halfDT;
				yVelocity += yAcceleration * halfDT;
			}
			
			// Set display position and scaling based on zoom
			sprite.scale = GameWorld.zoom * spriteScale;
			x = xPosition * GameWorld.zoom;
			y = yPosition * GameWorld.zoom;
			
			super.update();
		}
		
		/**
		 * Calculate current acceleration due to gravity
		 */
		private function CalculateGravity():void
		{
			// Sun exerts gravity
			var gravityVector:Point = GravityVector( xPosition, yPosition, 0, 0, Sun.MASS );
			xAcceleration = gravityVector.x;
			yAcceleration = gravityVector.y;
			
			// Planets exert gravity if flag is set
			if ( affectedByPlanets )
			{
				// Get all planets
				var planets:Vector.<Planet> = new Vector.<Planet>();
				world.getClass( Planet, planets );
				
				for each( var planet:Planet in planets )
				{
					if ( planet != this && planet.exertsGravity )
					{
						gravityVector = GravityVector( xPosition, yPosition, planet.xPosition, planet.yPosition, planet.radius );
						xAcceleration += gravityVector.x;
						yAcceleration += gravityVector.y;
					}
				}
			}
		}
		
		/**
		 * Calculate the vector of acceleration due to force from gravity
		 * exerted by object 2 on object 1
		 * 
		 * @param	x1
		 * Position of object 1
		 * @param	y1
		 * Position of object 1
		 * @param	x2
		 * Position of object 2
		 * @param	y2
		 * Position of object 2
		 * @param	mass
		 * Mass of object 2
		 * @return
		 * Vector of acceleration due to gravity on object 1
		 */
		private function GravityVector( x1:Number, y1:Number, x2:Number, y2:Number, mass:Number ):Point
		{
			// Calculate acceleration due to gravity
			// Force due to gravity = G * m1 * m2 / distance ^ 2
			// Acceleration due to force = f / m1
			var deltaX:Number = x2 - x1;
			var deltaY:Number = y2 - y1;
			var distanceSq:Number = deltaX * deltaX + deltaY * deltaY;
			var acceleration:Number = GRAVITATIONAL_CONSTANT * mass / distanceSq;
			
			// Make acceleration vector in the direction of object 2
			var distance:Number = Math.sqrt(distanceSq);
			var normal:Number = acceleration / distance;
			var gravityVector:Point = new Point(deltaX * normal, deltaY * normal);
			
			return gravityVector;
		}
		
		/**
		 * Draw a permanent tracer on the screen to trace the orbital path
		 */
		private function LeaveTracer():void
		{
			var tracer:Image = Image.createCircle( TRACER_RADIUS );
			tracer.centerOrigin();
			world.addGraphic( tracer, 0, x, y );
		}
		
		/**
		 * Calculate the stable starting orbital velocity for the planet and
		 * place the planet at a random point on that orbit.
		 * 
		 * @param	distance
		 * Distance from the sun
		 */
		private function CircularVelocity( distance:Number, setAngle:Number ):void
		{
			var cos:Number = Math.cos( setAngle );
			var sin:Number = Math.sin( setAngle );
			
			xPosition = cos * distance;
			yPosition = sin * distance;
			
			var velocity:Number = Math.sqrt( GRAVITATIONAL_CONSTANT * Sun.MASS / distance );
			
			xVelocity = sin * velocity;
			yVelocity = -cos * velocity;
		}
		
		/**
		 * Override this function to respond to collisions
		 */
		public function CollisionResponse( body:CelestialBody ):void
		{
		}
	}

}