package  
{
	import net.flashpunk.graphics.Stamp;
	import net.flashpunk.graphics.Text;
	import net.flashpunk.graphics.TiledImage;
	import net.flashpunk.tweens.misc.Alarm;
	import net.flashpunk.tweens.misc.VarTween;
	import net.flashpunk.World;
	import net.flashpunk.FP;
	import net.flashpunk.utils.*;
	
	/**
	 * Main class implementing the game
	 * 
	 * @author Switchbreak
	 */
	public class GameWorld extends World 
	{
		// Embeds
		[Embed(source = "../img/Nebula.png")] private static const NEBULA_BG:Class;
		[Embed(source = "../fonts/Dosis-Bold.ttf", embedAsCFF = 'false', fontFamily = 'UI Font')] public static const GAME_FONT:Class;
		
		// Constants
		static private const ASTEROID_COUNT:int = 8;
		static private const SLOW_MOTION:Number = 0.1;
		static private const ASTEROID_CLICK_THRESHOLD:Number = 200;
		static private const LINE_COLOR:uint = 0xFFFFFF;
		
		// Game states
		static private const STATE_NOTHING:int = 0;
		static private const STATE_THROWING:int = 1;
		static private const STATE_FOLLOWING:int = 2;
		static private const STATE_ENDSCREEN:int = 3;
		
		// Game objects
		public var sun:Sun;
		private var selectedAsteroid:Asteroid = null;
		private var scoreDisplay:Text;
		private var multiplierDisplay:Text;
		private var completionLabel:Text;
		private var completion:Text;
		private var nebula:TiledImage;
		private var endScreenLabel:Text;
		private var replayButton:Text;
		
		// Statics
		public static var gameState:int = STATE_NOTHING;
		public static var zoom:Number = 0.25;
		public static var timeSpeed:Number = 1;
		public static var score:Number = 0;
		public static var multiplier:Number = 0;
		
		/**
		 * Create and initialize the game world
		 */
		public function GameWorld() 
		{
			super();
			
			FP.randomizeSeed();
			//FP.console.debug = true;
			//FP.console.enable();
			
			gameState = STATE_NOTHING;
			zoom = 0.25;
			timeSpeed = 1;
			score = 0;
			multiplier = 0;
			
			// Set game font
			Text.font = "UI Font";
			
			// Center the scene on origin 0, 0
			camera.x = -FP.halfWidth;
			camera.y = -FP.halfHeight;
			
			// Create game objects
			sun = new Sun();
			add( sun );
			
			for ( var i:int = 0; i < ASTEROID_COUNT; i++ )
				add( new Asteroid( i / ASTEROID_COUNT ) );
			
			add( new Planet( 400, 40, 0x00FF77, 100 ) );
			add( new Planet( 700, 60, 0x00FF77 ) );
			add( new Planet( 1000, 80, 0x00FF77 ) );
			add( new Planet( 1400, 100, 0x00FF77 ) );
			add( new Planet( 1800, 100, 0x00FF77 ) );
			add( new Planet( 2500, 200, 0x00FF77 ) );
			add( new Planet( 3000, 200, 0x00FF77 ) );
			add( new Planet( 3500, 200, 0x00FF77 ) );
			
			//addTween( new Alarm( 1, SpawnUFO, LOOPING ), true );
			
			addGraphic( new Starfield(), 1000 );
			
			scoreDisplay = new Text( "Score: " + Math.floor( score ).toString() );
			scoreDisplay.scrollX = 0;
			scoreDisplay.scrollY = 0;
			addGraphic( scoreDisplay, -1000 );
			multiplierDisplay = new Text( Math.floor( multiplier ).toString() + "x", 0, scoreDisplay.height );
			multiplierDisplay.scrollX = 0;
			multiplierDisplay.scrollY = 0;
			addGraphic( multiplierDisplay, -1000 );
			
			completionLabel = new Text( "Completion:", 0, multiplierDisplay.y + multiplierDisplay.height );
			completionLabel.scrollX = 0;
			completionLabel.scrollY = 0;
			addGraphic( completionLabel, -1000 );
			completion = new Text( "0%", 0, completionLabel.y + completionLabel.height );
			completion.scrollX = 0;
			completion.scrollY = 0;
			addGraphic( completion, -1000 );
			
			nebula = new TiledImage( NEBULA_BG, 1280, 960 );
			nebula.scrollX = 0;
			nebula.scrollY = 0;
			nebula.setOffset( -(camera.x + FP.halfWidth) / GameWorld.zoom, -(camera.y + FP.halfHeight) / GameWorld.zoom );
			addGraphic( nebula, 1000 );
			nebula.scale = GameWorld.zoom * 10;
			nebula.x = FP.halfWidth - 6400 * GameWorld.zoom;
			nebula.y = FP.halfHeight - 4800 * GameWorld.zoom;
		}
		
		// Handle input
		override public function update():void 
		{
			// Zoom with mouse wheel input
			if ( Input.mouseWheel )
			{
				zoom += (Input.mouseWheelDelta * 0.005);
			}
			if ( Input.check( Key.NUMPAD_ADD ) || Input.check( Key.PAGE_DOWN ) )
			{
				zoom += 0.005;
			}
			if ( Input.check( Key.NUMPAD_SUBTRACT ) || Input.check( Key.PAGE_UP ) )
			{
				zoom -= 0.005;
			}
			if ( zoom < 0.1 )
				zoom = 0.1;
			if ( zoom > 1.5 )
				zoom = 1.5;
			
			// Handle input for different game states
			switch( gameState )
			{
				case STATE_NOTHING:
					// Check for clicks on asteroids to start throwing them
					if ( Input.mousePressed )
					{
						var asteroids:Vector.<Asteroid> = new Vector.<Asteroid>();
						getClass( Asteroid, asteroids );
						
						// Check if the click is near any asteroids
						for each( var asteroid:Asteroid in asteroids )
						{
							if ( (asteroid.x - mouseX) * (asteroid.x - mouseX) + (asteroid.y - mouseY) * (asteroid.y - mouseY) < ASTEROID_CLICK_THRESHOLD )
							{
								// Click was near enough to asteroid, select it
								// and go into slow motion until the mouse
								// button is released
								selectedAsteroid = asteroid;
								timeSpeed = SLOW_MOTION;
								gameState = STATE_THROWING;
								break;
							}
						}
					}
					
					break;
				case STATE_THROWING:
					// Throw asteroid when mouse is released
					if ( Input.mouseReleased )
					{
						if ( selectedAsteroid != null )
							selectedAsteroid.Throw( mouseX - selectedAsteroid.x, mouseY - selectedAsteroid.y );
						
						timeSpeed = 1.0;
						var zoomTween:VarTween = new VarTween(null, ONESHOT);
						zoomTween.tween( GameWorld, "zoom", 0.5, 0.1 );
						addTween( zoomTween, true );
						gameState = STATE_FOLLOWING;
					}
					
					break;
				case STATE_ENDSCREEN:
					if ( Input.mousePressed &&
						Input.mouseX >= replayButton.x && Input.mouseX <= replayButton.x + replayButton.width &&
						Input.mouseY >= replayButton.y && Input.mouseY <= replayButton.y + replayButton.height )
					{
						FP.world = new StartScreen();
					}
			}
			
			super.update();
			
			// Set follow camera post-update to avoid jitters from zoom change
			if ( gameState == STATE_FOLLOWING && selectedAsteroid != null )
			{
				camera.x = selectedAsteroid.x - FP.halfWidth;
				camera.y = selectedAsteroid.y - FP.halfHeight;
			}
			
			nebula.setOffset( -(camera.x + FP.halfWidth) / (GameWorld.zoom * 10), -(camera.y + FP.halfHeight) / (GameWorld.zoom * 10) );
			nebula.scale = GameWorld.zoom * 10;
			nebula.x = FP.halfWidth - 6400 * GameWorld.zoom;
			nebula.y = FP.halfHeight - 4800 * GameWorld.zoom;
			
			// Update completion percentage display
			var planets:Vector.<Planet> = new Vector.<Planet>();
			getClass( Planet, planets );
			
			var totalEvolution:Number = 0;
			var evolution:Number = 0;
			for each( var planet:Planet in planets )
			{
				totalEvolution += Planet.MAX_EVOLUTION;
				evolution += planet.evolution;
			}
			
			if( totalEvolution > 0 )
				completion.text = Math.round( (evolution / totalEvolution) * 100 ).toString() + "%";
			
			// Check for win/lose condition
			if ( gameState != STATE_ENDSCREEN )
			{
				if ( evolution >= totalEvolution )
					ShowEndScreen( "All Planets Full of Life" );
				
				if ( selectedAsteroid != null )
					evolution += selectedAsteroid.bacteriaLevel;
				if ( evolution <= 0 )
					ShowEndScreen( "All Life Has Ended" );
			}
			
			scoreDisplay.text = "Score: " + Math.floor( score ).toString();
			multiplierDisplay.text = Math.floor( multiplier ).toString() + "x"
		}
		
		/**
		 * Draw game graphics not handled by the engine
		 */
		override public function render():void 
		{
			super.render();
			
			// Draw a line to the selected asteroid, if needed
			if ( gameState == STATE_THROWING && selectedAsteroid != null )
			{
				Draw.linePlus( mouseX, mouseY, selectedAsteroid.x, selectedAsteroid.y, LINE_COLOR );
			}
		}
		
		/**
		 * Removes an asteroid from the world, handling game state change if the
		 * asteroid was selected for the current game state
		 * 
		 * @param	asteroid
		 * Asteroid to remove
		 */
		public function RemoveAsteroid( asteroid:Asteroid ):void
		{
			if ( selectedAsteroid == asteroid )
			{
				selectedAsteroid = null;
				if ( gameState == STATE_FOLLOWING )
				{
					camera.x = -FP.halfWidth;
					camera.y = -FP.halfHeight;
					var zoomTween:VarTween = new VarTween(null, ONESHOT);
					zoomTween.tween( GameWorld, "zoom", 0.25, 0.1 );
					addTween( zoomTween, true );
				}
				
				gameState = STATE_NOTHING;
			}
			
			remove( asteroid );
		}
		
		private function ShowEndScreen( endText:String ):void 
		{
			if ( selectedAsteroid != null )
			{
				selectedAsteroid.Explode();
				selectedAsteroid = null;
			}
			
			gameState = STATE_ENDSCREEN;
			
			endScreenLabel = new Text( endText );
			endScreenLabel.size = 60;
			endScreenLabel.y = (FP.height >> 2) - (endScreenLabel.height >> 1);
			endScreenLabel.x = 0;
			endScreenLabel.width = FP.width;
			endScreenLabel.align = "center";
			endScreenLabel.scrollX = 0;
			endScreenLabel.scrollY = 0;
			addGraphic( endScreenLabel, -2000 );
			
			replayButton = new Text( "Replay" );
			replayButton.size = 36;
			replayButton.y = FP.halfHeight + (FP.height >> 2) - (replayButton.height >> 1);
			replayButton.x = FP.halfWidth - (replayButton.width >> 1);
			replayButton.align = "center";
			replayButton.scrollX = 0;
			replayButton.scrollY = 0;
			addGraphic( replayButton, -2000 );
		}
		
		private function SpawnUFO():void
		{
			add( new UFO( -5000, 5000 ) );
		}
	}

}