package
{
	import flash.display.BitmapData;
	import flash.display.BlendMode;
	import flash.geom.Rectangle;
	import net.flashpunk.Entity;
	import net.flashpunk.graphics.Image;
	import net.flashpunk.graphics.Graphiclist;
	import net.flashpunk.FP;
	import net.flashpunk.Sfx;
	import net.flashpunk.tweens.misc.Alarm;
	import net.flashpunk.utils.Draw;
	
	/**
	 * Planet game object
	 *
	 * @author Switchbreak
	 */
	public class Planet extends CelestialBody
	{
		// Embeds
		[Embed(source = "../img/PlanetHealthBars.png")] static private const HEALTH_BAR:Class;
		[Embed(source = "../img/PlanetShading.png")] static private const SHADING:Class;
		[Embed(source = "../snd/Explosion30.mp3")] static private const EXPLODE:Class;
		
		// Constants
		static public const MAX_HEALTH:Number = 100;
		static public const MAX_EVOLUTION:Number = 200;
		static private const MAX_HEALTH_RATE:Number = 5;
		static private const MAX_EVOLUTION_RATE:Number = 3;
		static private const ASTEROID_HEALTH_COST:Number = 20;
		static private const RAIN_RATE:Number = 1;
		
		static private const EXPLODE_PARTICLE_COUNT:int = 10;
		static private const EXPLODE_PARTICLE_SPEED:int = 25;
		static private const EXPLODE_PARTICLE_SIZE:int = 30;
		static private const EXPLODE_PARTICLE_DURATION:Number = 2;
		static private const HEALTH_BAR_SPACING:Number = 25;
		static private const HURT_COLOR:uint = 0x00CC00;
		static private const HURT_LIFELESS_COLOR:uint = 0x888888;
		static private const LIFELESS_COLOR:uint = 0xCCCCCC;
		static private const HURT_MAXLIFE_COLOR:uint = 0x000088;
		static private const MAXLIFE_COLOR:uint = 0x00CCFF;
		static private const SOUND_EFFECTS_VOLUME:Number = 0.5;
		
		// Private variables
		public var health:Number;
		public var evolution:Number;
		public var showMode:Boolean = false;
		private var color:uint;
		private var healthBarScale:Number;
		private var shadingScale:Number;
		
		// Graphics
		private var planetSprite:Image;
		private var indicator:Indicator;
		private var particleEmitter:Graphiclist;
		private var healthBar:Image;
		private var evolveBar:Image;
		private var shading:Image;
		
		// Objects
		private var healthMask:BitmapData;
		private var evolveMask:BitmapData;
		
		/**
		 * Create and initialize the planet
		 * 
		 * @param	distance
		 * Distance from the sun
		 * @param	setRadius
		 * Size of the planet
		 * @param	setColor
		 * Color of the planet
		 */
		public function Planet(distance:Number, setRadius:Number, setColor:uint, startingEvolution:Number = 0, setShowMode:Boolean = false)
		{
			health = MAX_HEALTH;
			evolution = startingEvolution;
			color = setColor;
			showMode = setShowMode;
			
			planetSprite = new Image( SHADING );
			planetSprite.smooth = true;
			
			super( distance, setRadius, planetSprite, FP.random * Math.PI * 2, setRadius / (planetSprite.width >> 1) );
			
			// Create indicator object
			indicator = new Indicator( planetSprite.color );
			indicator.visible = false;
			addGraphic( indicator );
			
			if ( !showMode )
			{
				// Create health bar
				healthBar = new Image( HEALTH_BAR, new Rectangle( 0, 0, 125, 250 ) );
				healthMask = new BitmapData( healthBar.width, healthBar.height, true, 0 );
				healthBar.originX = healthBar.width;
				healthBar.originY = healthBar.height >> 1;
				healthBar.smooth = true;
				addGraphic( healthBar );
				
				// Create evolution bar
				evolveBar = new Image( HEALTH_BAR, new Rectangle( 125, 0, 125, 250 ) );
				evolveMask = new BitmapData( evolveBar.width, evolveBar.height, true, 0 );
				evolveBar.originX = 0;
				evolveBar.originY = healthBar.height >> 1;
				evolveBar.smooth = true;
				addGraphic( evolveBar );
				
				// Setup scaling on bars
				healthBarScale = radius / (healthBar.width - HEALTH_BAR_SPACING);
				var barScale:Number = GameWorld.zoom * healthBarScale;
				healthBar.scale = barScale;
				evolveBar.scale = barScale;
				
				UpdateHealthBar();
				UpdateEvolutionBar();
			}
		}
		
		/**
		 * Handle showing the indicator if offscreen
		 */
		override public function update():void 
		{
			super.update();
			
			if ( !showMode )
			{
				// Check if the planet is off the current screen
				var screenSize:Number = radius * GameWorld.zoom;
				setHitbox( screenSize * 2, screenSize * 2, screenSize, screenSize );
				if ( !onCamera )
				{
					indicator.PointAt( x, y );
					indicator.visible = true;
				}
				else
				{
					indicator.visible = false;
				}
				
				// Recharge health and evolution bar
				if ( health < MAX_HEALTH )
				{
					health += MAX_HEALTH_RATE * FP.elapsed;
					if ( health >= MAX_HEALTH )
					{
						health = MAX_HEALTH;
					}
					
					UpdateHealthBar();
				}
				
				if ( evolution > 0 && evolution < MAX_EVOLUTION )
				{
					evolution += MAX_EVOLUTION_RATE * (evolution / MAX_EVOLUTION) * FP.elapsed;
					if ( evolution > MAX_EVOLUTION )
						evolution = MAX_EVOLUTION;
					UpdateEvolutionBar();
				}
				
				var barScale:Number = GameWorld.zoom * healthBarScale;
				healthBar.scale = barScale;
				evolveBar.scale = barScale;
			}
			
			planetSprite.angle = FP.angle( xPosition, yPosition, 0, 0 );
		}
		
		override public function render():void 
		{
			super.render();
		}
		
		/**
		 * Respond to collisions from asteroids
		 */
		override public function CollisionResponse( body:CelestialBody ):void 
		{
			if ( body is Asteroid )
			{
				health -= ASTEROID_HEALTH_COST;
				UpdateHealthBar();
			}
			
			// Show hit particle effect
			if ( particleEmitter == null )
			{
				particleEmitter = new Graphiclist();
				particleEmitter.active = true;
				addGraphic( particleEmitter );
			}
			
			var xVector:Number = body.xPosition - xPosition;
			var yVector:Number = body.yPosition - yPosition;
			var distance:Number = Math.sqrt( xVector * xVector + yVector * yVector );
			xVector /= distance;
			yVector /= distance;
			
			for ( var i:int; i < EXPLODE_PARTICLE_COUNT; ++i )
			{
				particleEmitter.add( new Particle( xPosition + xVector * radius, yPosition + yVector * radius, 10, color, FP.random * 100 - 50, FP.random * 100 - 50, EXPLODE_PARTICLE_DURATION, RemoveParticle ) );
			}
			
			if ( health <= 0 )
				Explode();
		}
		
		public function Rain( rainAmount:Number ):void
		{
			evolution += rainAmount * RAIN_RATE;
			UpdateEvolutionBar();
		}
		
		public function Gather():Number
		{
			var amountGathered:Number = ASTEROID_HEALTH_COST;
			if ( amountGathered > evolution )
				amountGathered = evolution;
			
			evolution -= amountGathered;
			UpdateEvolutionBar();
			
			return amountGathered;
		}
		
		/**
		 * Remove a previously emitted particle
		 * 
		 * @param	particle
		 * Particle to remove
		 */
		private function RemoveParticle( particle:Particle ):void
		{
			particleEmitter.remove( particle );
		}
		
		/**
		 * Explode the asteroid
		 */
		public function Explode():void
		{
			var sound:Sfx = new Sfx( EXPLODE );
			sound.play( SOUND_EFFECTS_VOLUME );
			
			collidable = false;
			moving = false;
			exertsGravity = false;
			planetSprite.visible = false;
			healthBar.visible = false;
			evolveBar.visible = false;
			
			if ( particleEmitter == null )
			{
				particleEmitter = new Graphiclist();
				particleEmitter.active = true;
				addGraphic( particleEmitter );
			}
			
			for ( var i:int; i < EXPLODE_PARTICLE_COUNT; ++i )
			{
				var xVector:Number = FP.random * EXPLODE_PARTICLE_SPEED * 2 - EXPLODE_PARTICLE_SPEED;
				var yVector:Number = FP.random * EXPLODE_PARTICLE_SPEED * 2 - EXPLODE_PARTICLE_SPEED;
				particleEmitter.add( new Particle( xPosition, yPosition, EXPLODE_PARTICLE_SIZE, color, xVector, yVector, EXPLODE_PARTICLE_DURATION, RemoveParticle ) );
			}
			
			addTween( new Alarm( EXPLODE_PARTICLE_DURATION, RemoveSelf, ONESHOT ), true );
		}
		
		public function RemoveSelf():void
		{
			world.remove( this );
		}
		
		private function UpdateHealthBar():void 
		{
			var maskHeight:Number = health * healthBar.height / MAX_HEALTH;
			
			healthMask.fillRect( new Rectangle( 0, 0, healthBar.width, healthBar.height - maskHeight ), 0 );
			healthMask.fillRect( new Rectangle( 0, healthBar.height - maskHeight, healthBar.width, maskHeight ), 0xFFFFFFFF );
			healthBar.drawMask = healthMask;
			
			UpdatePlanetColor();
		}
		
		private function UpdateEvolutionBar():void
		{
			var maskHeight:Number = evolution * evolveBar.height / MAX_EVOLUTION;
			
			evolveMask.fillRect( new Rectangle( 0, 0, evolveBar.width, evolveBar.height - maskHeight ), 0 );
			evolveMask.fillRect( new Rectangle( 0, evolveBar.height - maskHeight, evolveBar.width, maskHeight ), 0xFFFFFFFF );
			evolveBar.drawMask = evolveMask;
			
			UpdatePlanetColor();
		}
		
		private function UpdatePlanetColor():void
		{
			var planetColor:uint;
			
			if ( evolution >= MAX_EVOLUTION )
			{
				if( health < MAX_HEALTH )
					planetColor = HURT_MAXLIFE_COLOR;
				else
					planetColor = MAXLIFE_COLOR;
			}
			else if ( evolution > 0 )
			{
				if( health < MAX_HEALTH )
					planetColor = HURT_COLOR;
				else
					planetColor = color;
			}
			else
			{
				if( health < MAX_HEALTH )
					planetColor = HURT_LIFELESS_COLOR;
				else
					planetColor = LIFELESS_COLOR;
			}
			
			planetSprite.color = planetColor;
			indicator.color = planetColor;
		}
	}

}